/*
 * This module can be used to trigger a double fault by overflowing
 * the kernel stack.
 *
 * kldload kstackovf
 * sysctl debug.kstackovf=1
 * - recursively call a function to overflow stack
 * sysctl debug.kstackovf=2
 * - push the LAPIC version register on the stack at a 16-bit operand
 * sysctl debug.kstackovf=3
 * - push the LAPIC version register on the stack as a 32-bit operand (i386)
 *   or a 64-bit operand (amd64).
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/kthread.h>

#include <x86/apicreg.h>

extern lapic_t *lapic;
extern void pushmmio(void);
extern void pushmmio16(void);

int *stackaddr[KSTACK_PAGES * PAGE_SIZE];

static void
recurse(int iter)
{
	int i;
	stackaddr[iter] = &i;
	recurse(iter + 1);
}

static int
sysctl_doit(SYSCTL_HANDLER_ARGS)
{
	register_t rflags;
	int error, doit;

	doit = 0;
	error = sysctl_handle_int(oidp, &doit, 0, req);
	if (error || req->newptr == NULL || doit == 0)
		return (error);

	rflags = intr_disable();
	if (doit == 1)
		recurse(1);
	else if (doit == 2)
		pushmmio16();
	else
		pushmmio();
	intr_restore(rflags);

	return (0);
}

SYSCTL_PROC(_debug, OID_AUTO, kstackovf, CTLTYPE_INT|CTLFLAG_RW, 0, 0, 
    sysctl_doit, "I", "");

static int
kstackovf_modevent(module_t mod, int type, void *unused)
{

        switch (type) {
        case MOD_LOAD:
		break;
	case MOD_UNLOAD:
		break;
	default:
		return (EINVAL);
	}

	return (0);
}

static moduledata_t kstackovf_mod_data= {
	"kstackovf",
	kstackovf_modevent,
};

DECLARE_MODULE(test, kstackovf_mod_data, SI_SUB_SMP + 1, SI_ORDER_ANY);
